"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.HMSSessionStore = void 0;
var _reactNative = require("react-native");
var _HMSManagerModule = _interopRequireDefault(require("../modules/HMSManagerModule"));
var _HMSConstants = require("./HMSConstants");
var _HMSLogger = require("./HMSLogger");
var _HMSUpdateListenerActions = require("./HMSUpdateListenerActions");
var _utils = require("../utils");
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
function _defineProperty(e, r, t) { return (r = _toPropertyKey(r)) in e ? Object.defineProperty(e, r, { value: t, enumerable: !0, configurable: !0, writable: !0 }) : e[r] = t, e; }
function _toPropertyKey(t) { var i = _toPrimitive(t, "string"); return "symbol" == typeof i ? i : i + ""; }
function _toPrimitive(t, r) { if ("object" != typeof t || !t) return t; var e = t[Symbol.toPrimitive]; if (void 0 !== e) { var i = e.call(t, r || "default"); if ("object" != typeof i) return i; throw new TypeError("@@toPrimitive must return a primitive value."); } return ("string" === r ? String : Number)(t); }
/**
 * Session store is a shared realtime key-value store that is accessible by all Peers in a Room.
 * It can be utilized to implement features such as pinned text, spotlight (which brings a particular
 * peer to the center stage for everyone in the room) and more, enhancing interactive experiences.
 *
 * To get an instance of `HMSSessionStore` class, You can add an event listener for `ON_SESSION_STORE_AVAILABLE`
 * event on the `HMSSDK` instance
 *
 * For example:
 * ```
 * hmsInstance.addEventListener(HMSUpdateListenerActions.ON_SESSION_STORE_AVAILABLE, <your callback function>);
 * ```
 *
 * Checkout Session Store docs fore more details ${@link https://www.100ms.live/docs/react-native/v2/how-to-guides/interact-with-room/room/session-store}
 */
class HMSSessionStore {
  constructor() {
    _defineProperty(this, "_deviceEventEmitterSubscription", void 0);
    _defineProperty(this, "_eventEmitter", void 0);
    _defineProperty(this, "_addedKeyChangeListenerCount", 0);
  }
  /**
   * This method sets a value for a specific key on session store.
   * Once a value is assigned, it will be available for other peers in the room
   * who are listening for changes in value for that specific key.
   *
   * @param {JsonValue} value - The value to set, which can be any JSON-compatible type.
   * @param {string} key - The key under which to store the value.
   * @returns {Promise} A promise that resolves when the operation is complete.
   */
  async set(value, key) {
    const data = await _HMSManagerModule.default.setSessionMetadataForKey({
      id: _HMSConstants.HMSConstants.DEFAULT_SDK_ID,
      key,
      value
    });
    return data;
  }

  /**
   * Retrieves the value for a given key from the session store. This method does not subscribe
   * to changes; it only returns the current value at the time of the call.
   *
   * To listen to value change updates use `addKeyChangeListener` method instead.
   *
   * @param {string} key - The key whose value is to be retrieved.
   * @returns {Promise} A promise that resolves with the value of the key.
   */
  async get(key) {
    const data = await _HMSManagerModule.default.getSessionMetadataForKey({
      id: _HMSConstants.HMSConstants.DEFAULT_SDK_ID,
      key
    });
    return data;
  }

  /**
   * Registers a callback to listen for changes to specified keys in the session store.
   * The callback is called with the initial value and again whenever any value changes.
   *
   * @param {string[]} forKeys - An array of keys to listen for changes on.
   * @param {Function} callback - The function to call when a value changes.
   * @returns {Object} An object with a `remove` method to unregister the listener.
   */
  addKeyChangeListener(forKeys, callback) {
    // Add Native Device Event Emitter if it is not already added
    if (!this._deviceEventEmitterSubscription) {
      this._deviceEventEmitterSubscription = _reactNative.DeviceEventEmitter.addListener(_HMSUpdateListenerActions.HMSUpdateListenerActions.ON_SESSION_STORE_CHANGED, this._deviceEventEmitterListener.bind(this));
    }

    // Create JS side EventEmitter
    if (!this._eventEmitter) {
      this._eventEmitter = new _utils.EventEmitter();
    }

    // Unique Identifier for adding native event listener
    const uniqueId = forKeys.join('') + '_' + Date.now().toString();
    const eventEmitter = this._eventEmitter;

    // Add listeners on eventEmitter for each key
    const subscriptions = forKeys.map(key => eventEmitter.addListener(key, callback, {
      uniqueId
    }));

    //
    let cleanupHandler = () => {
      this._cleanup(subscriptions);
    };

    // Adding 'KeyChangeListener' on native side
    _HMSManagerModule.default.addKeyChangeListener({
      id: _HMSConstants.HMSConstants.DEFAULT_SDK_ID,
      keys: forKeys,
      uniqueId
    })
    // Adding 'KeyChangeListener' fails on native side
    .catch(err => {
      if (typeof cleanupHandler === 'function') {
        callback(err, null);
        cleanupHandler();
        cleanupHandler = null;
      }
    });
    this._addedKeyChangeListenerCount += 1;
    return {
      remove: () => {
        if (typeof cleanupHandler === 'function') {
          cleanupHandler();
          cleanupHandler = null;
        }
      }
    };
  }
  _cleanup(subscriptionsToRemove) {
    // Extracting `uniqueId` from first subscription (all subscriptions have same uniqueId)
    // this `uniqueId` will be used to remove 'KeyChangeListener' from native side
    const uniqueId = subscriptionsToRemove.length > 0 ? subscriptionsToRemove[0].context.uniqueId : null;

    // Removing required subscriptions from 'eventEmitter'
    subscriptionsToRemove.forEach(subscription => subscription.remove());

    // Removing 'KeyChangeListener' from native side
    _HMSManagerModule.default.removeKeyChangeListener({
      id: _HMSConstants.HMSConstants.DEFAULT_SDK_ID,
      uniqueId
    }).catch(error => {
      const logger = (0, _HMSLogger.getLogger)();
      logger === null || logger === void 0 || logger.verbose("Error while removing key change listener, Listener didn't get registerred at first place or was already removed", error);
    });
    this._addedKeyChangeListenerCount -= 1;
    if (this._addedKeyChangeListenerCount <= 0) {
      this._addedKeyChangeListenerCount = 0;
      if (this._deviceEventEmitterSubscription && typeof this._deviceEventEmitterSubscription.remove === 'function') {
        this._deviceEventEmitterSubscription.remove();
      } else {
        _reactNative.DeviceEventEmitter.removeListener(_HMSUpdateListenerActions.HMSUpdateListenerActions.ON_SESSION_STORE_CHANGED, this._deviceEventEmitterListener);
      }
      this._deviceEventEmitterSubscription = undefined;
      this._eventEmitter = undefined;
    }
  }
  _deviceEventEmitterListener(data) {
    var _getLogger, _this$_eventEmitter;
    // if id is different from default sdk_id, return early
    if (data.id !== _HMSConstants.HMSConstants.DEFAULT_SDK_ID) {
      return;
    }

    // emit event for the key
    (_getLogger = (0, _HMSLogger.getLogger)()) === null || _getLogger === void 0 || _getLogger.verbose('#Listener ON_SESSION_STORE_CHANGED event: ', data);
    (_this$_eventEmitter = this._eventEmitter) === null || _this$_eventEmitter === void 0 || _this$_eventEmitter.emit(data.key, null, data);
  }
}
exports.HMSSessionStore = HMSSessionStore;
//# sourceMappingURL=HMSSessionStore.js.map