"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.HmsViewComponent = void 0;
var _react = _interopRequireWildcard(require("react"));
var _reactNative = require("react-native");
var _HMSConstants = require("./HMSConstants");
var _HMSVideoViewMode = require("./HMSVideoViewMode");
var _hmsviews = require("../hooks/hmsviews");
function _getRequireWildcardCache(e) { if ("function" != typeof WeakMap) return null; var r = new WeakMap(), t = new WeakMap(); return (_getRequireWildcardCache = function (e) { return e ? t : r; })(e); }
function _interopRequireWildcard(e, r) { if (!r && e && e.__esModule) return e; if (null === e || "object" != typeof e && "function" != typeof e) return { default: e }; var t = _getRequireWildcardCache(r); if (t && t.has(e)) return t.get(e); var n = { __proto__: null }, a = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var u in e) if ("default" !== u && {}.hasOwnProperty.call(e, u)) { var i = a ? Object.getOwnPropertyDescriptor(e, u) : null; i && (i.get || i.set) ? Object.defineProperty(n, u, i) : n[u] = e[u]; } return n.default = e, t && t.set(e, n), n; }
/**
 * Interface defining the properties for the `HmsView` component.
 *
 * This interface specifies the structure of the props that the `HmsView` component expects. It includes
 * properties for configuring the video track display, such as the track ID, mirroring options,
 * and scale type. It also includes properties for handling events and customizing the component's style.
 *
 * @interface HmsViewProps
 * @property {Object} data - An object containing the track ID, instance ID, mirroring option, and scale type for the video or audio track.
 * @property {string} data.trackId - The unique identifier for the track to be displayed.
 * @property {string} data.id - The identifier for the `HmsViewComponent` instance.
 * @property {boolean} data.mirror - Indicates whether the video should be mirrored. This is commonly used for local video tracks.
 * @property {HMSVideoViewMode} data.scaleType - Determines how the video fits within the bounds of the view (e.g., aspect fill, aspect fit).
 * @property {boolean} autoSimulcast - Enables automatic simulcast layer switching based on network conditions, if supported.
 * @property {boolean} setZOrderMediaOverlay - When true, the video view will be rendered above the regular view hierarchy.
 * @property {ViewStyle} style - Custom styles to apply to the view.
 * @property {Function} onChange - A callback function that is invoked when the `HmsView` component emits a change event.
 * @property {Function} onDataReturned - A callback function that is invoked when the `HmsView` component returns data in response to a capture frame event.
 *
 * @see {https://www.100ms.live/docs/react-native/v2/how-to-guides/set-up-video-conferencing/render-video/overview}
 */

// Imports the `HmsView` component from the native side using the `requireNativeComponent` function.
// This component is used to render video tracks in the application.
const HmsView = (0, _reactNative.requireNativeComponent)('HMSView');
let _nextRequestId = 1;
let _requestMap = new Map();

/**
 * Defines the properties for the `HmsViewComponent`.
 *
 * This interface outlines the props that can be passed to the `HmsViewComponent` to configure its behavior and appearance.
 *
 * @interface HmsComponentProps
 * @property {string} trackId - The unique identifier for the track to be displayed.
 * @property {ViewStyle} [style] - Optional. Custom styles to apply to the view.
 * @property {boolean} [mirror] - Optional. If true, the video will be mirrored. This is commonly used for local video tracks.
 * @property {boolean} [autoSimulcast] - Optional. Enables automatic simulcast layer switching based on network conditions, if supported.
 * @property {HMSVideoViewMode} [scaleType] - Optional. Determines how the video fits within the bounds of the view (e.g., aspect fill, aspect fit).
 * @property {boolean} [setZOrderMediaOverlay] - Optional. When true, the video view will be rendered above the regular view hierarchy.
 * @property {string} id - The identifier for the `HmsViewComponent` instance.
 *
 * @see {https://www.100ms.live/docs/react-native/v2/how-to-guides/set-up-video-conferencing/render-video/overview}
 */

const HmsViewComponent = exports.HmsViewComponent = /*#__PURE__*/_react.default.forwardRef((props, ref) => {
  const {
    trackId,
    style = styles.hmsView,
    id = _HMSConstants.HMSConstants.DEFAULT_SDK_ID,
    mirror = false,
    setZOrderMediaOverlay = false,
    autoSimulcast = true,
    scaleType = _HMSVideoViewMode.HMSVideoViewMode.ASPECT_FILL
  } = props;
  const hmsViewRef = (0, _react.useRef)();
  const [applyStyles_ANDROID, setApplyStyles_ANDROID] = (0, _react.useState)(false);
  const data = {
    trackId,
    id,
    mirror,
    scaleType
  };

  /**
   * This method is passed to `onChange` prop of `HmsView` Native Component.
   * It is invoked when `HmsView` emits 'topChange' event.
   */
  const onChange = ({
    nativeEvent
  }) => {
    const {
      event,
      data
    } = nativeEvent;
    setApplyStyles_ANDROID(true);
    if (event === 'ON_RESOLUTION_CHANGE_EVENT') {
      (0, _hmsviews.setHmsViewsResolutionsState)(trackId, data);
    }
  };

  /**
   * This method is passed to `onDataReturned` prop of `HmsView` Native Component.
   * It is invoked when `HmsView` emits 'captureFrame' event.
   */
  const _onDataReturned = event => {
    // We grab the relevant data out of our event.
    let {
      requestId,
      result,
      error
    } = event.nativeEvent;
    // Then we get the promise we saved earlier for the given request ID.
    let promise = _requestMap.get(requestId);
    if (result) {
      // If it was successful, we resolve the promise.
      promise.resolve(result);
    } else {
      // Otherwise, we reject it.
      promise.reject(error);
    }
    // Finally, we clean up our request map.
    _requestMap.delete(requestId);
  };
  const capture = async () => {
    const viewManagerConfig = _reactNative.UIManager.getViewManagerConfig('HMSView');
    let requestId = _nextRequestId++;
    let requestMap = _requestMap;

    // We create a promise here that will be resolved once `_onRequestDone` is
    // called.
    let promise = new Promise(function (resolve, reject) {
      requestMap.set(requestId, {
        resolve,
        reject
      });
    });
    _reactNative.UIManager.dispatchViewManagerCommand((0, _reactNative.findNodeHandle)(hmsViewRef.current), viewManagerConfig.Commands.capture, [requestId]);
    return promise;
  };
  (0, _react.useImperativeHandle)(ref, () => {
    return {
      capture
    };
  });
  return /*#__PURE__*/_react.default.createElement(HmsView, {
    ref: hmsViewRef,
    onChange: onChange,
    data: data,
    style: _reactNative.Platform.OS === 'android' ? applyStyles_ANDROID ? style : {} : style,
    autoSimulcast: autoSimulcast,
    scaleType: scaleType,
    setZOrderMediaOverlay: setZOrderMediaOverlay,
    onDataReturned: _onDataReturned
  });
});
const styles = _reactNative.StyleSheet.create({
  hmsView: {
    flex: 1
  }
});
//# sourceMappingURL=HmsView.js.map